"""This module is used for calculating the MFCC feature vectors for the input speech utterance. The following sequence of steps are followed:

1. Pre-emphasizing input utterance by passing it through a first order HPF
2. Dividing utterance into non-overlapping frames
3. Computing DFT of each frame
4. Passing each frame through the Mel filterbank and calculating the energy passed by each filter
5. Computing logarithm of filterbank energies
6. Computing DCT of log filterbank energies and extracting first 13 coefficients
7. Appending spectral dynamic features to give a 39 dimensional feature vector for each frame"""
import numpy as np
from scipy.fftpack import dct
import time
import timeit

class MFCC:
	"""
	This class contains all methods and parameters related to MFCC calculation.

	:param alpha: Parameter used in pre-emphasis filtering. Should lie between 0 and 1
	:type alpha: float
	:param N: Number of points used for computing DFT
	:type N: int
	:param fs: Sampling frequency
	:type fs: int
	:param frame_dur: Duration of 1 frame in seconds
	:type frame_due: float

	Documentation related to all methods of this class is described below.
	
	"""
	def __init__(self, alpha=0.95, N=256, fs=8000, eps=10e-6, frame_dur=0.025):
		self.alpha = alpha   #Preemphasis parameter
		self.N = N	     # No of FFT points	
		self.eps = eps       # eps to avoid mathematical errors
		self.fs = fs
		self.mfcc_frame_length = round(fs*frame_dur)  #frame length for mfcc
		self.prev_frame = np.zeros((14,))
		self.prev_prev_frame = np.zeros((14,))

	def generate_filter_bank(self, num_filters=23, lower_freq=300, upper_freq=3800):
		"""
		This function computes the Mel filterbank. The filters are stored in rows, the columns correspond
    		to DFT bins. The filters are returned as a matrix of size nfilt * (nfft/2 + 1)

		:param num_filters: Number of filters in the Mel filterbank
		:type num_filters: int
		:param lower_freq: Lower frequency from where the first filter should start
		:type lower_freq: int
		:param upper_freq: Upper frequency where the last filter should end. Should be less than (fs/2)
		:type upper_freq: int
		:returns: Matrix containing filterbank weights
		:rtype: matrix

    		"""
    		self.highfreq = upper_freq
    		assert self.highfreq <= self.fs/2, "highfreq is greater than samplerate/2"
    
    		# compute points evenly spaced in mels
    		self.lowmel = self.hz2mel(lower_freq)
    		self.highmel = self.hz2mel(upper_freq)
    		self.melpoints = np.linspace(self.lowmel,self.highmel,num_filters+2)
    		self.bin = np.floor((self.N+1)*self.mel2hz(self.melpoints)/self.fs)

		self.fbank = np.zeros([num_filters,self.N/2+1])
    		for j in range(0,num_filters):
        		for i in range(int(self.bin[j]), int(self.bin[j+1])):
            			self.fbank[j,i] = (i - self.bin[j]) / (self.bin[j+1]-self.bin[j])
        		for i in range(int(self.bin[j+1]), int(self.bin[j+2])):
            			self.fbank[j,i] = (self.bin[j+2]-i) / (self.bin[j+2]-self.bin[j+1])
		return self.fbank
		
		
	def hz2mel(self, hz):
		"""
		Function for converting frequency in Hertz to Mel scale
		"""
    		return 2595 * np.log10(1+hz/700.0)
    
	def mel2hz(self, mel):
		"""
		Function for converting frequency in Mel scale to Hertz		
		"""
    		return 700*(10**(mel/2595.0)-1)
	
	def compute_mfcc(self, input_frame, include_dc=False):
		"""
		This function performs all the steps as mentioned earlier inorder to calculate 39 dimensional MFCC feature vector for an input frame.
	
		:param input_frame: Frame of audio from which MFCC features have to be extracted
		:type input_frame: array
		:param include_dc: Used to indicate whether or not the 1st MFCC coefficient is to be included
		:type include_dc: boolean
		:returns: 39 dimensional MFCC vector
		:rtype: array
		"""
		self.windowed_frame = input_frame * np.hamming(len(input_frame)) #Hamming Windowing input frame
		self.pre_frame = self.windowed_frame - self.alpha*np.concatenate((np.full(1, 0), self.windowed_frame[:-1])) #Preemphasis
		self.fft_frame = np.fft.rfft(self.pre_frame,self.N) #Computing FFT
		self.mel_fb_sq_weights = self.generate_filter_bank()**2
		self.mel_frame = np.log(np.dot(self.mel_fb_sq_weights,(np.abs(self.fft_frame)**2)) + self.eps)  #Passing through mel and taking log
		self.mel_frame_normalized = self.mel_frame / np.sum(self.mel_fb_sq_weights, axis=1)     #normalizing by filter energies
		self.dct_frame = dct(self.mel_frame_normalized)     #Computing DCT 
		self.delta = self.dct_frame[0:14] - self.prev_frame
		self.double_delta = self.dct_frame[0:14] - self.prev_prev_frame
		self.coeffs = np.append(self.dct_frame[0:14], [self.delta[1:14], self.double_delta[1:14]])
		self.prev_prev_frame = self.prev_frame
		self.prev_frame = self.dct_frame[0:14]
		if (include_dc):
			return self.coeffs  	# Returning 39 MFCC coefficients
		else:
			return self.coeffs[1:40]




"""This module is used for recording speech utterances both for training as well as testing purposes. PyAudio is required for this module."""
import pyaudio
from array import array
import wave,sys
from struct import pack
class wordRecorder:
	"""

	This class contains methods for recording audio from a microphone, segmenting audio to remove noise by using an amplitude based VAD and for storing segmented utterances in an appropriate folder.

	:param samplingFrequency: Frequency at which we want to sample audio
	:type samplingFrequency: int
	:param threshold: Threshold used for amplitude based VAD (scaled in the range 0-16384)
	:type threshold: int

	Documentation related to all member functions is listed below.

	"""
	def __init__(self, samplingFrequency = 8000, threshold = 14000):
		self.samplingFrequency = samplingFrequency
		self.threshold = threshold
	
	def isSilent(self, data):
		"""

		This function is used to check if the whole recorded audio is silence or not. If it is silence, then this utterance is discarded.

		:param data: Recorded audio 
		:type data: array
		:returns: 1 if entire audio is silence and 0 otherwise
		:rtype: boolean

		"""
	        return max(data) < self.threshold

	def normalize(self, data):
		"""

		This function is used to normalize the sampled audio stream such that all values lie in the range -16383 to 16384. This is because we use a 16-bit representation to store audio. Out of these 16 bits 1 bit is reserved as a sign bit.

		:param data: Recorded audio
		:type data: array
		:returns: Normalized audio
		:rtype: array

		"""
        	maxShort = 16384
    		scale = float(maxShort)/max(abs(i) for i in data)

		r = array('h')
    		for i in data:
        		r.append(int(i*scale))
    		return r

	def trimWord(self, data):
		"""
		
		This function implements the amplitude based Voice Activity detector. It segments out audio based on whether the amplitude of the audio is greater than the specified threshold or not.

		:param data: Normalized audio
		:type data: array
		:returns: Trimmed audio containing only speech segments
		:rtype: array
		
		"""
    		def trimStart(data):
        		snd_started = False
        		r = array('h')

        		for i in data:
				if not snd_started and abs(i)>self.threshold:
                			snd_started = True
                			r.append(i)

            			elif snd_started:
                			r.append(i)
        		return r

		data = trimStart(data)
		data.reverse()
		data = trimStart(data)
		data.reverse()
		return data

	def record(self):
		"""

		This function implements the recording routine used for getting audio from a microphone using PyAudio. It also calls the ``normalize()`` and ``trimWord()`` methods to return the normalized and trimmed audio containing speech only.

		:returns: Trimmed and normalized recorded audio
		:rtype: array

		"""
    		p = pyaudio.PyAudio()
		stream = p.open(format=pyaudio.paInt16, channels=1, rate=self.samplingFrequency, input=True, output=False, frames_per_buffer=1024)

		num_silent = 0
		snd_started = False
		
		r = array('h')

		for i in range(int(self.samplingFrequency*2/1024)):
        		snd_data = array('h', stream.read(1024))
        		r.extend(snd_data)
		
		sample_width = p.get_sample_size(pyaudio.paInt16)
    		stream.stop_stream()
		stream.close()
		p.terminate()

    		r = self.normalize(r)
    		r = self.trimWord(r)
    		return sample_width, r

	def record2File(self, path):
		"""

		This function is used to store the recorded audio after it has been normalized and trimmed into a specified directory as a .wav file.

		:param path: Path to directory where audio is to be stored
                :type data: str
                
		"""
    		sample_width, data = self.record()
    		data = pack('<' + ('h'*len(data)), *data)

		wf = wave.open(path, 'wb')
		wf.setnchannels(1)
		wf.setsampwidth(sample_width)
		wf.setframerate(self.samplingFrequency)
		wf.writeframes(data)
		wf.close()




"""This module detects the trained hotword using the mfcc,dtw and wordrecorder classes. The following sequence of steps are followed:

1. Compute MFCC feature vectors of the test utterance
2. Compute DTW distance between the test utterance and each of the 10 stored training utterances using these MFCC vectors
3. Compute the mean of these DTW distances
4. If the mean of these distances is greater than some threshold then no hotword is present otherwise hotword is present"""
import numpy as np
import os
import scipy.io.wavfile as wv
import mfcc
import dtw

class hwDetector:
	"""

	This class contains methods which accomplish the end goal of hotword detection using methods defined in mfcc,dtw and wordrecorder classes.

	:param samplingFrequency: Sampling frequency of audio
	:type samplingFrequency: int
	:param framePeriod: Duration of 1 frame in seconds
	:type framePeriod: float
	:param hopPeriod: Frame hopping duration in seconds
	:type hopPeriod: float
	:param trainDir: Path to directory containing training utterances
	:type trainDir: str
	:param thresh: Threshold used for DTW distance 
	:type thresh: float

	Documentation related to all methods in this class is described below.

	"""
	def __init__(self, samplingFrequency = 8000, framePeriod = 25e-3, hopPeriod = 10e-3, trainDir="./train_audio/",thresh=6):
		self.samplingFrequency = samplingFrequency
		self.framePeriod = framePeriod
		self.hopPeriod = hopPeriod
		self.trainDir = trainDir
		self.thresh = thresh
		self.hopLength = int(samplingFrequency * hopPeriod)
		self.frameLength = int(samplingFrequency * framePeriod)
		self.referenceMFCC = []
		for file_name in os.listdir(trainDir):
			if file_name.endswith(".wav"):
				(fs, data) = wv.read(trainDir + file_name)
				num_frames = int(data.shape[0]/self.hopLength) - int(np.ceil(self.frameLength/self.hopLength))

				MFCC_calculator = mfcc.MFCC()
				MFCC_MATRIX = np.empty([39, num_frames])
				for k in range(num_frames):
					MFCC_MATRIX[:,k] = MFCC_calculator.compute_mfcc(data[k*self.hopLength : k*self.hopLength + self.frameLength])
				self.referenceMFCC.append(MFCC_MATRIX)

		DTW_calculator = dtw.DTW()
		distance_list = [DTW_calculator.compute_distance(np.transpose(matrix),np.transpose(matrix2)) for matrix in self.referenceMFCC for matrix2 in self.referenceMFCC]
		self.thresh = np.mean(np.array((distance_list)))*1.2

	def distance(self, fileName):
		"""
	
		This function is used for calculating the DTW distance between the test utterance and each of the 10 training utterances.

		:param fileName: Name of test utterance .wav file
		:type fileName: str
		:returns: List of DTW distances of test utterance with each training utterance
		:rtype: list

		"""
		DTW_calculator = dtw.DTW()
		(fs, data) = wv.read(fileName)
		num_frames = int(data.shape[0]/self.hopLength) - int(np.ceil(self.frameLength/self.hopLength))
		if num_frames<=0:
			return 10000

		MFCC_calculator = mfcc.MFCC()
		MFCC_MATRIX = np.empty([39, num_frames])
		for k in range(num_frames):
			MFCC_MATRIX[:,k] = MFCC_calculator.compute_mfcc(data[k*self.hopLength : k*self.hopLength + self.frameLength])

		distance_list = [DTW_calculator.compute_distance(np.transpose(matrix),np.transpose(MFCC_MATRIX)) for matrix in self.referenceMFCC]
		return distance_list

	def isHotword(self,fileName):
		"""
		
		This function computes the mean of the calculated DTW distances and returns a True value if mean distance is less than a specified threshold. If mean distance is greater than threshold it returns False indicating absence of hotword.

		:param fileName: Name of test utterance .wav file
                :type fileName: str
                :returns: True/False to indicate presence/absence of hotword
                :rtype: Boolean

		"""
		distances = np.array(self.distance(fileName))
		mean_dist = np.mean(distances)
		if (mean_dist < self.thresh):
			return True
		return False




"""This module is used for computing the DTW distance between two utterances."""
import numpy as np
import math

def euclideanDistance(a,b):
	"""

	This function computes the simple Euclidean distance between two MFCC vectors. This is further useful in DTW distance computation.

	:param a: MFCC vector 1
	:type a: array
	:param b: MFCC vector 2
	:type b: array
	:returns: Euclidean distance between MFCC vectors 1 and 2
	:rtype: float

	"""
        return np.sqrt(np.abs(np.sum(np.dot(a,a) - np.dot(b,b))))

class DTW:
	"""
	
	This class contains methods and parameters used for computing the DTW distance between training and test utterances. 

	Documentation related to all methods in this class is described below.

	"""
        def __init__(self, distFunc=euclideanDistance):
                self.distFunc = distFunc 

        def compute_distance(self, reference, test):
		"""
		This function computes the DTW distance between two utterances by constructing a distance matrix.

		:param reference: Matrix containing MFCC vectors of training utterance
		:type reference: matrix
		:param test: Matrix containing MFCC vectors of test utterance
		:param test: matrix
		:returns: DTW distance between reference and test utterance
		:rtype: float
 
		"""
                if (type(reference) is str):
	                reference = list(reference)
        	if (type(test) is str):
                	test = list(test)
	        if (type(reference) is list):
	                reference = np.array(reference)
	        if (type(test) is list):
	                test = np.array(test)
	        DTW_matrix = np.empty([reference.shape[0],test.shape[0]])
	        DTW_matrix[:] = np.inf

	        DTW_matrix[0,0] = 0

	        for i in range(reference.shape[0]):
	                for j in range(test.shape[0]):
	                        cost=euclideanDistance(reference[i,:],test[j,:])
	                        r_index = i-1
	                        c_index = j-1
	                        if(r_index < 0):
	                                r_index = 0
	                        if(c_index < 0):
	                                c_index = 0
	                        DTW_matrix[i,j] = cost + min(DTW_matrix[r_index,j],DTW_matrix[i,c_index],DTW_matrix[r_index,c_index])
	        return DTW_matrix[-1,-1]/(test.shape[0]+reference.shape[0])



import sys
import os

sys.path.append(os.path.abspath('./hotword_detection'))
import wordRecorder as wr
import hwDetector as hd

wrdRec = wr.wordRecorder()
hwDet = hd.hwDetector()

print("Speak a word")
wrdRec.record2File("demo.wav")
print(hwDet.isHotword("demo.wav"))



from setuptools import setup #enables develop

setup(name='hotword_detection',
      version='1.0',
      description='Python Hotword Detection',
      author='Prakhar Swarup, Saketh Sharma, Hitesh Tulsiani',
      author_email='prkhrswrp@gmail.com',
      license='IITB',
      scripts=['bin/checkHotword.py','bin/trainHotword.py'],
      url='https://github.com/sakethgsharma/HotWordDetection',
      packages=['hotword_detection'],
      install_requires=[
        "numpy",
        "scipy",
        "wave",
        "pyaudio",
      ],
)



import sys
import os

if not (os.path.exists('./train_audio')):
	os.mkdir('./train_audio')
sys.path.append(os.path.abspath('./hotword_detection'))
import wordRecorder as wr
import time

wRec = wr.wordRecorder()
filelist = [ f for f in os.listdir("./train_audio/") if f.endswith(".wav") ]
for f in filelist:
    os.remove("./train_audio/" + f)
print("Record hotword instances...")
for i in range(10):
	print("Sample" + str(i))
	wRec.record2File("./train_audio/" + str(int(time.time())) + ".wav")

